// ========================================================================
// matrix4x4.h
//
// R^(4x4)
// ========================================================================


#ifndef	__MATRIX4X4_H
#define	__MATRIX4X4_H


#include "vector4.h"
#include "matrix3x3.h"

#undef  _MATRIX_ROWMAX
#define	_MATRIX_ROWMAX	4

//
//
class matrix4x4
{
friend vector3;
friend vector4;
friend class matrix3x3;
private:
	vector4 e[4];

public:
	matrix4x4() {}
	matrix4x4(
		VAR   e11,VAR   e12,VAR   e13,VAR   e14,
	    VAR   e21,VAR   e22,VAR   e23,VAR   e24,
		VAR   e31,VAR   e32,VAR   e33,VAR   e34,
		VAR   e41,VAR   e42,VAR   e43,VAR   e44
		){
		e[0]=vector4(e11,e21,e31,e41);
		e[1]=vector4(e12,e22,e32,e42);
		e[2]=vector4(e13,e23,e33,e43);
		e[3]=vector4(e14,e24,e34,e44);
	}
	matrix4x4(const matrix4x4& a);
	matrix4x4(const VAR   a);
	matrix4x4(vector4& e1,vector4& e2,vector4& e3,vector4& e4){
		e[0]=e1;e[1]=e2;e[2]=e3;e[3]=e4;
	}
	matrix4x4(const matrix3x3& m);

	~matrix4x4() {}

	void clear();

	// one term operator
	vector4& operator[](int n){ return e[n]; }
	vector4 row(int n) { return e[n]; }
	vector4 col(int n) { return vector4(e[0][n],e[1][n],e[2][n],e[3][n]); }
	VAR  & operator()(int i,int j){ return e[j][i]; }


	// two term operator
	friend matrix4x4 operator+(matrix4x4& a,matrix4x4& b);
	friend matrix4x4 operator-(matrix4x4& a,matrix4x4& b);
	friend matrix4x4 operator*(matrix4x4& a,matrix4x4& b);
	friend matrix4x4 operator/(matrix4x4& a,matrix4x4& b);

	friend vector3 operator*(matrix4x4& a,vector3&   b);

	matrix4x4& operator+=(matrix4x4& a);
	matrix4x4& operator-=(matrix4x4& a);
	matrix4x4& operator*=(matrix4x4& a);
	matrix4x4& operator/=(matrix4x4& a);


	//
	matrix4x4 inv();
	matrix4x4 transpose();

	//
	void SetMove(vector3& v);
	bool SetView(vector3& from,vector3& at,vector3& up);
	bool SetProj(VAR   fov,VAR   aspect,VAR   nearplane,VAR   farplane);
};

inline void matrix4x4::clear(){
	*this=0.0f;
}


inline matrix4x4::matrix4x4(const matrix4x4& a){
	for(int i=_MATRIX_ROWMAX-1;i>=0;i--)
		e[i]=a.e[i];
}


inline matrix4x4::matrix4x4(const VAR   a){
	e[0]=vector4(a,0,0,0);
	e[1]=vector4(0,a,0,0);
	e[2]=vector4(0,0,a,0);
	e[3]=vector4(0,0,0,a);
}

inline matrix4x4::matrix4x4(const matrix3x3& m){
	e[0]=vector4(m[0],0);
	e[1]=vector4(m[1],0);
	e[2]=vector4(m[2],0);
	e[3]=vector4(0,0,0,1);
}

// ========================================================================


inline matrix4x4 operator+(matrix4x4& a,matrix4x4& b){
	return matrix4x4(
		a.e[0]+b.e[0],a.e[1]+b.e[1],a.e[2]+b.e[2],a.e[3]+b.e[3]
		);
}

inline matrix4x4 operator-(matrix4x4& a,matrix4x4& b){
	return matrix4x4(
		a.e[0]-b.e[0],a.e[1]-b.e[1],a.e[2]-b.e[2],a.e[3]-b.e[3]
		);
}

inline matrix4x4 operator*(matrix4x4& a,matrix4x4& b)
{
	matrix4x4 c;
	for(int i=_MATRIX_ROWMAX;--i>=0;)
	{
		for(int j=_MATRIX_ROWMAX;--j>=0;)
		{
			VAR   f=0;
			for(int k=_MATRIX_ROWMAX;--k>=0;)
				f+=a.e[k][i]*b.e[j][k];
			c.e[j][i]=f;
		}
	}
	return c;
}

inline matrix4x4 operator/(matrix4x4& a,matrix4x4& b){
	return a*b.inv();
}

inline vector3   operator*(matrix4x4& a,vector3&  v){
	return vector3( a[0][0]*v[0]+a[1][0]*v[1]+a[2][0]*v[2] + a[3][0] ,
	                a[0][1]*v[0]+a[1][1]*v[1]+a[2][1]*v[2] + a[3][1] ,
	                a[0][2]*v[0]+a[1][2]*v[1]+a[2][2]*v[2] + a[3][2] );
}


inline matrix4x4 matrix4x4::inv()
{
	matrix4x4 a,b;
	a=*this;
	b=1.0f;
	for(int i=0;i<_MATRIX_ROWMAX;i++)
	{
		int u=i;
		for(int k=i+1;k<_MATRIX_ROWMAX;k++)
		{
			if( fabs(a.e[k][i])>fabs(a.e[u][i]) )
				u=k;
		}
		std::swap(a.e[u],a.e[i]);
		std::swap(b.e[u],b.e[i]);

		b.e[i]/=a.e[i][i];
		a.e[i]/=a.e[i][i];
		for(int j=0;j<_MATRIX_ROWMAX;j++)
		{
			if(i==j) continue;
			VAR   d=(a.e[j][i]/a.e[i][i]);
			a.e[j]-=a.e[i]* d;
			b.e[j]-=b.e[i]* d;
		}
	}

	return b;
}


// ========================================================================

inline matrix4x4& matrix4x4::operator+=(matrix4x4& a){
	return *this=*this+a;
}

inline matrix4x4& matrix4x4::operator-=(matrix4x4& a){
	return *this=*this-a;
}

inline matrix4x4& matrix4x4::operator*=(matrix4x4& a){
	return *this=*this*a;
}

inline matrix4x4& matrix4x4::operator/=(matrix4x4& a){
	return *this=*this/a;
}

// ========================================================================

inline matrix4x4 matrix4x4::transpose(){
	return matrix4x4(
		e[0][0],e[0][1],e[0][2],e[0][3],
		e[1][0],e[1][1],e[1][2],e[1][3],
		e[2][0],e[2][1],e[2][2],e[2][3],
		e[3][0],e[3][1],e[3][2],e[3][3]
		);
}

// ========================================================================


inline void matrix4x4::SetMove(vector3& v){
	*this=matrix4x4(
		1.0f,0.0f,0.0f, v[0],
		0.0f,1.0f,0.0f, v[1],
		0.0f,0.0f,1.0f, v[2],
		0.0f,0.0f,0.0f, 1.0f
		);
}


#endif //__MATRIX4X4_H
