// ========================================================================
// quaternion.h
//
// quaternion
// ========================================================================


#ifndef	__QUATERNION_H
#define	__QUATERNION_H

#include "vector3.h"
#include "matrix4x4.h"

class quaternion
{
private:
	VAR   s;
	vector3 v;

public:
	quaternion() : s(0.0f),v(0.0f) {}
	quaternion(vector3& _v) : s(0.0f),v(_v) {}
	quaternion(const VAR _s) : s(_s),v(0.0f) {}
	quaternion(const VAR _s,const VAR _x,const VAR _y,const VAR _z){ s=_s;v=vector3(_x,_y,_z); }
	quaternion(const VAR _s,const vector3& _v) : s(_s),v(_v) {}
	quaternion(const quaternion& q) : s(q.s),v(q.v) {}

	~quaternion() {}

	// one term operator
	//quaternion& operator=(const VAR f) { return *this=quaternion(f); }
	//quaternion& operator=(const quaternion& q) { return *this=quaternion(q); }
	quaternion& operator-() { return *this=quaternion(-s,-v); }

	// two term operator
	quaternion& operator+=(quaternion& q);
	quaternion& operator-=(quaternion& q);
	quaternion& operator*=(quaternion& q);
	quaternion& operator*=(VAR       f);
	quaternion& operator/=(quaternion& q);

	friend quaternion operator+ (quaternion& a,quaternion& b);
	friend quaternion operator- (quaternion& a,quaternion& b);
	friend quaternion operator* (quaternion& a,quaternion& b);
	friend quaternion operator* (vector3&    a,quaternion& b);
	friend quaternion operator* (quaternion& a,vector3&    b);
	friend quaternion operator* (quaternion& a,VAR       b);
	friend quaternion operator* (VAR       a,quaternion  b);
	friend quaternion operator/ (quaternion& a,quaternion& b);
	friend quaternion operator/ (quaternion& a,VAR       b);

	// compare operator
	friend bool operator==(quaternion& a,quaternion& b);
	friend bool operator==(quaternion& a,VAR       b);
	friend bool operator==(VAR       a,quaternion& b);
	friend bool operator!=(quaternion& a,quaternion& b);
	friend bool operator!=(quaternion& a,VAR       b);
	friend bool operator!=(VAR       a,quaternion& b);


	// transform
	VAR norm();
	VAR abs();
	quaternion unit();
	quaternion conjugate();
	quaternion inv();
	matrix3x3 matrix();
	matrix3x3 rotation();

	// self transform
	void normalize();

	//
	void setrotation(vector3& _v,VAR theta); // rotation q from unit vector and radian

	//
	static quaternion rotation(vector3& _v,VAR theta); // rotation q from unit vector and radian
};

// ========================================================================

inline quaternion& quaternion::operator+=(quaternion& q){
	return *this=*this+q;
}

inline quaternion& quaternion::operator-=(quaternion& q){
	return *this=*this-q;
}

inline quaternion& quaternion::operator*=(quaternion& q){
	return *this=*this*q;
}

inline quaternion& quaternion::operator*=(VAR       f){
	return *this=*this*f;
}

inline quaternion& quaternion::operator/=(quaternion& q){
	return *this=*this/q;
}


// ========================================================================

inline quaternion operator+ (quaternion& a,quaternion& b){
	return quaternion(a.s+b.s,a.v+b.v);
}

inline quaternion operator- (quaternion& a,quaternion& b){
	return a+(-b);
}

inline quaternion operator* (quaternion& a,quaternion& b){
	return quaternion(
		a.s*b.s - a.v*b.v ,
		a.s*b.v + b.s*a.v + a.v.cross(b.v)
		);
}

inline quaternion operator* (vector3&    a,quaternion& b){
	return quaternion(a)*b;
}

inline quaternion operator* (quaternion& a,vector3&    b){
	return a*quaternion(b);
}

inline quaternion operator* (quaternion& a,VAR       b){
	return quaternion(a.s*b,a.v*b);
}

inline quaternion operator* (VAR       a,quaternion& b){
	return b*a;
}

inline quaternion operator/ (quaternion& a,quaternion& b){
	return a*b.inv();
}

inline quaternion operator/ (quaternion& a,VAR       b){
	return a*(1.0f/b);
}

// ========================================================================

inline bool operator==(quaternion& a,quaternion& b){
	return a.s==b.s && a.v==b.v;
}

inline bool operator==(quaternion& a,VAR       b){
	return a==quaternion(b);
}

inline bool operator==(VAR       a,quaternion& b){
	return quaternion(a)==b;
}

inline bool operator!=(quaternion& a,quaternion& b){
	return ! (a==b);
}

inline bool operator!=(quaternion& a,VAR       b){
	return ! (a==b);
}

inline bool operator!=(VAR       a,quaternion& b){
	return ! (a==b);
}

// ========================================================================

inline VAR quaternion::norm(){
	return s*s + v*v;
}

inline VAR quaternion::abs(){
	return sqrt(norm());
}

inline quaternion quaternion::unit(){
	VAR f=1.0f/abs();
	return quaternion(s*f,v*f);
}

inline quaternion quaternion::conjugate(){
	return quaternion(s,-v);
}

inline quaternion quaternion::inv(){
	return conjugate()/norm();
}

inline matrix3x3 quaternion::matrix(){
	return rotation();
}

inline matrix3x3 quaternion::rotation(){
	return matrix3x3(
		1 - 2*v[1]*v[1] - 2*v[2]*v[2] ,     2*v[0]*v[1] - 2*   s*v[2] ,     2*v[0]*v[2] + 2*   s*v[1] ,
		    2*v[0]*v[1] + 2*   s*v[2] , 1 - 2*v[0]*v[0] - 2*v[2]*v[2] ,     2*v[1]*v[2] - 2*   s*v[0] ,
		    2*v[0]*v[2] - 2*   s*v[1] ,     2*v[1]*v[2] + 2*   s*v[0] , 1 - 2*v[0]*v[0] - 2*v[1]*v[1]
			);
}

inline void quaternion::normalize()
{
	*this=unit();
}

// ========================================================================

inline void quaternion::setrotation(vector3& _v,VAR theta)
{
	s=cos(theta/2);
	v=_v*sin(theta/2);
}

// ========================================================================

inline quaternion quaternion::rotation(vector3& _v,VAR theta)
{
	quaternion q;
	q.setrotation(_v,theta);
	return q;
}

#endif //__QUATERNION_H

